/*
 * TorusBackgroundGenerator.java
 *
 * Created on February 14, 2004, 2:42 PM
 */
package MySimulation;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Component;
import javax.swing.JPanel;
import java.awt.Dimension;
import java.util.Vector;
import java.util.*;


/**
 *Torus BackgroundGenerator
 *
 * this class generates a torus background based on algorithm descirbed by
 * Dr. Dave Mount.
 * Using the similar technique as KB tree, we will generate a background that is wrapped around
 * also, we will track all the segment as boundry and allow changing colors for them
 * @author  Tzu-hsiu Chiou
 */
public class TorusBackgroundGenerator {
    
    JPanel background = null;
    Vector segments = new Vector();
    Dimension r = null;
    int num = 0;
    int bkg[][];
    int x,y;
    int sx, sy;   // use to track split
    boolean swrap;
    int dir = 0;
    boolean wrap;
    /** Creates a new instance of TorusBackgroundGenerator */
    public TorusBackgroundGenerator(int inum, JPanel bk) {
        background = bk;
        r = bk.getSize();
        num = inum;
        generate();
    }
    
    public void generate() {
        segments.clear();
        int maxx = getMaxX();         // these are for boundries checking
        int maxy = getMaxY();
        
        int incx = 0, incy = 0;
        int counter = 0;
        boolean wrap;
        
        allocatesBackgroundG();  // allocates array thats' storing all information
        // also intialize all elements to -1
        
        // -1 : empty, allow ray to move through
        // -2 : joint, multiple ray intersects at this point
        // >= 0: the ray that's occuping current point
        
        for(int i = 0; i < num; i++) {
            
            // generates a random coordinate
            x = (int) (Math.random() * maxx);
            y = (int) (Math.random() * maxy);
            
            while(bkg[x][y] != -1)  // if the location is already taken
            {
                x = (int) (Math.random() * maxx);
                y = (int) (Math.random() * maxy);
            }
            
            
            //if ( (Math.random() < 0.5)) 
            if (i%2 == 1)
            {
                
                incx = 1;
                incy = 0;
                
            }
            else {incy = 1; incx = 0;}
            
            rayexpands(incx,incy);
            
        }
    }
    
    
    
    public void rayexpands(int incx, int incy)
    // this funciton takes a point and a direction, expands until runs into something, then create a line wher the points expands
    {
        int origx = x;
        int origy = y;
        int x1,x2,y1,y2;          // use to track where segment start and end
        // boolean wrap = false;     // track if wrap around occurs
        wrap = false;
        boolean owrap = wrap;
        
        x1 = x;
        x2 = x;
        y2 = y;
        y1 = y;
        
        int reservcounter = getCurrentCounter();
        segments.setSize(getCurrentCounter()+1);
        segments.setElementAt(null,reservcounter); // add new line
        
        // now move in "positive" direction i.e. x and y increases
        while(bkg[x][y] == -1) {
            bkg[x][y] = reservcounter;
            owrap = wrap;
            incPoint(incx,incy);
            
            
            if (bkg[x][y] != -1)  // if the location is not empty
            {
                //                  System.out.println("stop at:" + x + "," + y + " :" + bkg[x][y] + ";");
                //x1 = x -incx;
                //y1 = y -incy;
                incPoint(-incx,-incy);
                x1 = x;
                y1 = y;
                incPoint(incx,incy);
                
                wrap = owrap;
                if (bkg[x][y] != reservcounter) // collides with some other line
                {                 
                  
                    //bkg[x][y] = -2;     // to indicates a joint
                  splitLine(x, y,  incy, incx);   // incy and incx are reversed for split
                                                        
                }
            }
            
        }
         
        // move in negative direction
        x = origx;
        y = origy;
        incx = -incx;
        incy = -incy;
        incPoint(incx, incy);
        if (bkg[x][y] != -1)  // if the location is not empty back to original
        {
            incPoint(-incx,-incy);
            x2 = x ;
            y2 = y ;
            incPoint(incx,incy);
        }
        else{  // move to new point
            x2 =x;
            y2 =y;
        }
        
        
        while(bkg[x][y] == -1) {
            bkg[x][y] = reservcounter;
            owrap = wrap;
            incPoint(incx,incy);
            
            if (bkg[x][y] != -1)  // if the location is not empty
            {
                //                       System.out.println("stop2 at:" + x + "," + y + " :" + bkg[x][y] + ";");
                incPoint(-incx,-incy);
                x2 = x;
                y2 = y;               
                incPoint(incx,incy);
                 wrap = owrap;
                 
                if (bkg[x][y] != reservcounter) // collides with some other line
                {                    
                   // bkg[x][y] = -2;     // to indicates a joint
                   splitLine(x, y,  incy, incx);   // incy and incx are reversed for split
                }
            }
        }
         
            //  if (!((x1 == x2) && (y1 == y2))) {
        TorusLine temp = new TorusLine(x2, y2, x1, y1, wrap, Color.BLACK, r);
       // System.out.println("new lines: (" + origx + "," + origy + ")");
       // System.out.println("lines: (" + x2 + "," + y2 + "," + x1 + "," + y1 + ")");
        //TorusLine temp = new TorusLine(startx, starty, endx, endy, wrap, Color.green, r);
        //segments.setSize(getCurrentCounter()+1);
        segments.setElementAt(temp,reservcounter); // add new line
              //} else bkg[x1][y1] = -1;
        
        
        
    }
    
    public void splitLine(int isx, int isy, int sincx, int sincy)
    // this function look at the collision point and updates segments
    {
        
      if (bkg[isx][isy] == -2)  return;
        
        bkg[isx][isy] = -2;
        sincx = Math.abs(sincx);
        sincy = Math.abs(sincy);
        int posSegNum;  //  track the segment number for positive direction
        int negSegNum;  //  track the segment number for negative direction
        swrap = false;
        sx = isx;
        sy = isy;
        int x1, x2, y1, y2;
        boolean oswrap = swrap;
        
        if (bkg[sx][sy] != -2) throw new RuntimeException();
        
        // move in possitive direction
        
        incSplitPoint(sincx, sincy);
        swrap = false;
        x1 = x2 = sx;
        y1 = y2 = sy;
        posSegNum = bkg[sx][sy];
        
        
        // update upper segment
        if (posSegNum >= 0) {
            while(bkg[sx][sy] == posSegNum) {
                bkg[sx][sy] = getCurrentCounter();
                oswrap = swrap;
                incSplitPoint(sincx,sincy);
                
                
                if (bkg[sx][sy] != posSegNum)  // if the location is not same array
                {
                    incSplitPoint(-sincx, -sincy);    
                    x2 = sx;
                    y2 = sy;
                    swrap = oswrap;
                    
                }
            }
            
             if (!((x1 == x2) && (y1 == y2))) {
                TorusLine temp = new TorusLine(x2, y2, x1, y1, swrap, randColor(), r);
               
                segments.setSize(getCurrentCounter()+1);
                segments.setElementAt(temp,getCurrentCounter()-1); // add new line
                segments.setElementAt(null,posSegNum); // remove previous line
            } else bkg[x1][y1] = -2;
        }
       
        // updates lower segment
         swrap = false;
        sx = isx;
        sy = isy;
        if (bkg[sx][sy] != -2) throw new RuntimeException();
        sincx = -sincx;
        sincy = -sincy;            
        incSplitPoint(sincx, sincy);
        swrap = false;
        x1 = x2 = sx;
        y1 = y2 = sy;
        negSegNum = bkg[sx][sy];
        if (negSegNum >= 0) {
            while(bkg[sx][sy] == negSegNum) {
                bkg[sx][sy] = getCurrentCounter();
                
                oswrap = swrap;
                incSplitPoint(sincx,sincy);
                
                
                if (bkg[sx][sy] != negSegNum)  // if the location is not same ray anymore
                {
                    incSplitPoint(-sincx, -sincy);    
                    x2 = sx;
                    y2 = sy;
                    swrap = oswrap;   
                }
                
            }
            
            if (!((x1 == x2) && (y1 == y2))) {
                 
                TorusLine temp = new TorusLine(x2, y2, x1, y1, swrap, randColor(), r);
                segments.setSize(getCurrentCounter()+1);
                segments.setElementAt(temp,getCurrentCounter()-1); // add new line
                segments.setElementAt(null,negSegNum); // remove previous line                
            }
            else {bkg[x1][y1] = -2;}
        }
         
    }
    
    public Color randColor()
    {return new Color((int) (Math.random() * 255),(int) (Math.random() * 255),(int) (Math.random() * 255));}
    
    public void incPoint(int incx, int incy)
    // this function point point x in the direction specified (and wrap around if necessary
    {
        
        x  += incx;
        y  += incy;
        
        if (incx >= 0 && incy >= 0) {
            if (x >= getMaxX()) wrap = true;
            if (y >= getMaxY()) wrap = true;
            x %= getMaxX();
            y %= getMaxY();
        }
        else // wrap around in negative direction
        {
            if (x < 0) {x += getMaxX() ; wrap = true;}
            if (y < 0) {y += getMaxY() ; wrap = true;}
        }
    }
    
    public void incSplitPoint(int sincx, int sincy)
    // this function sx point  in the direction specified (and wrap around if necessary
    {
        
        sx  += sincx;
        sy  += sincy;
        
        if (sincx >= 0 && sincy >= 0) {
            if (sx >= getMaxX()) swrap = true;
            if (sy >= getMaxY()) swrap = true;
            sx %= getMaxX();
            sy %= getMaxY();
        }
        else // wrap around in negative direction
        {
            if (sx < 0) {sx += getMaxX() ; swrap = true;}
            if (sy < 0) {sy += getMaxY() ; swrap = true;}
        }
    }
    
    public boolean collides(int cx, int cy)
    {
        if (cx >= 0 && cy >= 0 && cx <getMaxX() && cy < getMaxY())          
        if (bkg[cx][cy] >= 0)
        {
            
            TorusLine temp = (TorusLine) segments.get(bkg[cx][cy]);
            if (temp == null) {System.out.println(bkg[cx][cy] + "at location" + cx + "," + cy + ")");  throw new RuntimeException();}
                temp.setFlash(50);
         return true;
        }
        
       
        return false;
    }
    
    public int getCurrentCounter()
    // this functions return the next location the a seg is going to be stored
    {return segments.size();}
    
    public void allocatesBackgroundG() {
        /// allocating space for the new array
        int maxx = getMaxX();
        int maxy = getMaxY();
        bkg = new int[maxx][];
        for(int i = 0; i< maxx; i++) {
            bkg[i] = new int[maxy];
        }
        for(int i = 0; i < maxx; i++)
            for (int j =0; j<maxy; j++)
                bkg[i][j] = -1;
        
    }
    
    public int getMaxX()
    {return (int) r.getWidth();}
    
    public int getMaxY()
    {return (int) r.getHeight();}
    
    
    public void paint(Graphics g) {
        for(Enumeration e = segments.elements(); e.hasMoreElements();) {
            TorusLine tempbh = (TorusLine) e.nextElement();
            if (tempbh != null)  tempbh.paint(g);
        }
    }
}
